"""Klasse `Connection`."""

from __future__ import annotations

__all__: Final[list[str]] = ["Connection"]

import socket
from contextlib import suppress
from typing import TYPE_CHECKING, Final

if TYPE_CHECKING:
    from io import TextIOWrapper


class Connection:
    """Objekte der Klasse `Connection` ermöglichen eine Netzwerkverbindung zu einem
    Server mittels TCP/IP-Protokoll. Nach Verbindungsaufbau können Zeichenketten
    (`Strings`) zum Server gesendet und von diesem empfangen werden. Zur
    Vereinfachung geschieht dies zeilenweise, d. h., beim Senden einer
    Zeichenkette wird ein Zeilentrenner ergänzt und beim Empfang wird dieser
    entfernt. Es findet nur eine rudimentäre Fehlerbehandlung statt, so dass z.B.
    der Zugriff auf unterbrochene oder bereits getrennte Verbindungen nicht zu
    einem Programmabbruch führt. Eine einmal getrennte Verbindung kann nicht
    reaktiviert werden.
    """

    __slots__: Final[tuple[str, str, str]] = ("_socket", "_to_server", "_from_server")

    def __init__(self, server_ip: str, server_port: int) -> None:
        """Ein Objekt vom Typ `Connection` wird erstellt. Dadurch wird eine Verbindung
        zum durch `server_ip` und `server_port` spezifizierten Server aufgebaut,
        so dass Daten (Zeichenketten) gesendet und empfangen werden können.
        Kann die Verbindung nicht hergestellt werden, kann die Instanz von Connection
        nicht mehr verwendet werden.
        """
        try:
            self._socket: socket.socket | None = socket.socket(
                socket.AF_INET,
                socket.SOCK_STREAM,
            )
            self._socket.connect((server_ip, server_port))
            self._to_server: TextIOWrapper | None = self._socket.makefile(
                mode="w",
                encoding="utf-8",
            )
            self._from_server: TextIOWrapper | None = self._socket.makefile(
                mode="r",
                encoding="utf-8",
            )
        except OSError:
            self._socket = None
            self._to_server = None
            self._from_server = None

    def receive(self) -> str | None:
        """Es wird beliebig lange auf eine eingehende Nachricht vom Server gewartet und
        diese Nachricht anschließend zurückgegeben. Der vom Server angehängte
        Zeilentrenner wird zuvor entfernt. Während des Wartens ist der ausführende
        Prozess blockiert. Wurde die Verbindung unterbrochen oder durch den Server
        unvermittelt geschlossen, wird `None` zurückgegeben.
        """
        if self._from_server is not None:
            with suppress(OSError, ValueError):
                received_line: str = self._from_server.readline().strip()
                return received_line if received_line else None
        return None

    def send(self, message: str) -> None:
        """Die Nachricht `message` wird - um einen Zeilentrenner ergänzt - an den Server
        gesendet. Schlägt der Versand fehl, geschieht nichts.
        """
        if self._to_server is not None:
            with suppress(OSError, ValueError):
                self._to_server.write(f"{message}\n")
                self._to_server.flush()

    def close(self) -> None:
        """Die Verbindung zum Server wird getrennt und kann nicht mehr verwendet werden.
        War die Verbindung bereits getrennt, geschieht nichts.
        """
        if self._socket is not None:
            with suppress(OSError):
                self._to_server.close()
                self._from_server.close()
                self._socket.shutdown(socket.SHUT_RDWR)
                self._socket.close()



# pylint: skip-file
__all__: Final[list[str]] = [
    "Stack",
    "Queue",
    "List",
    "BinaryTree",
    "ComparableContent",
    "ComparableContentT",
    "Vertex",
    "Edge",
    "BinarySearchTree",
    "Graph",
]

from typing import Final, Generic, TypeVar, overload

from nrw.datastructures._comparable_content import ComparableContent, ComparableContentT

_T = TypeVar("_T")

class Queue(Generic[_T]):
    __slots__: Final[tuple[str, str]] = ("_head", "_tail")
    __hash__ = None  # type: ignore[assignment]

    def __init__(self) -> None: ...
    @property
    def is_empty(self) -> bool: ...
    def enqueue(self, content: _T) -> None: ...
    def dequeue(self) -> None: ...
    @property
    def front(self) -> _T | None: ...

class Stack(Generic[_T]):
    __slots__: Final[tuple[str]] = ("_head",)
    __hash__ = None  # type: ignore[assignment]

    def __init__(self) -> None: ...
    @property
    def is_empty(self) -> bool: ...
    def push(self, content: _T) -> None: ...
    def pop(self) -> None: ...
    @property
    def top(self) -> _T | None: ...

class List(Generic[_T]):
    __slots__: Final[tuple[str, str, str]] = ("_first", "_last", "_current")
    __hash__ = None  # type: ignore[assignment]

    def __init__(self) -> None: ...
    @property
    def is_empty(self) -> bool: ...
    @property
    def has_access(self) -> bool: ...
    def next(self) -> None: ...
    def to_first(self) -> None: ...
    def to_last(self) -> None: ...
    @property
    def content(self) -> _T | None: ...
    @content.setter
    def content(self, new_content: _T | None) -> None: ...
    def insert(self, content: _T | None) -> None: ...
    def append(self, content: _T | None) -> None: ...
    def concat(self, other_list: List[_T] | None) -> None: ...
    def remove(self) -> None: ...

class BinaryTree(Generic[_T]):
    __slots__: Final[tuple[str]] = ("_node",)
    __hash__ = None  # type: ignore[assignment]

    @overload
    def __init__(self) -> None: ...
    @overload
    def __init__(
        self,
        content: _T | None,
    ) -> None: ...
    @overload
    def __init__(
        self,
        content: _T | None,
        left_tree: BinaryTree[_T] | None,
        right_tree: BinaryTree[_T] | None,
    ) -> None: ...
    @property
    def is_empty(self) -> bool: ...
    @property
    def content(self) -> _T | None: ...
    @content.setter
    def content(self, new_content: _T | None) -> None: ...
    @property
    def left_tree(self) -> BinaryTree[_T] | None: ...
    @left_tree.setter
    def left_tree(self, new_tree: BinaryTree[_T] | None) -> None: ...
    @property
    def right_tree(self) -> BinaryTree[_T] | None: ...
    @right_tree.setter
    def right_tree(self, new_tree: BinaryTree[_T] | None) -> None: ...

class BinarySearchTree(Generic[ComparableContentT]):
    __slots__: Final[tuple[str]] = ("_node",)
    __hash__ = None  # type: ignore[assignment]

    def __init__(self) -> None: ...
    @property
    def is_empty(self) -> bool: ...
    @property
    def content(self) -> ComparableContentT | None: ...
    @property
    def left_tree(self) -> BinarySearchTree[ComparableContentT] | None: ...
    @property
    def right_tree(self) -> BinarySearchTree[ComparableContentT] | None: ...
    def insert(self, content: ComparableContentT | None) -> None: ...
    def remove(self, content: ComparableContentT | None) -> None: ...
    def search(
        self,
        content: ComparableContentT | None,
    ) -> ComparableContentT | None: ...

class Vertex:
    __slots__: Final[tuple[str, str]] = ("_id", "_mark")
    __hash__ = None  # type: ignore[assignment]

    def __init__(self, id_: str) -> None: ...
    @property
    def id(self) -> str: ...
    @property
    def mark(self) -> bool: ...
    @mark.setter
    def mark(self, new_mark: bool) -> None: ...
    @property
    def is_marked(self) -> bool: ...

class Edge:
    __slots__: Final[tuple[str, str, str]] = ("_vertices", "_weight", "_mark")
    __hash__ = None  # type: ignore[assignment]

    def __init__(self, vertex: Vertex, another_vertex: Vertex, weight: int) -> None: ...
    @property
    def vertices(self) -> tuple[Vertex, Vertex]: ...
    @property
    def weight(self) -> int: ...
    @weight.setter
    def weight(self, new_weight: int) -> None: ...
    @property
    def mark(self) -> bool: ...
    @mark.setter
    def mark(self, new_mark: bool) -> None: ...
    @property
    def is_marked(self) -> bool: ...

class Graph:
    __slots__: Final[tuple[str, str]] = ("_vertices", "_edges")
    __hash__ = None  # type: ignore[assignment]

    def __init__(self) -> None: ...
    @property
    def vertices(self) -> List[Vertex]: ...
    @property
    def edges(self) -> List[Edge]: ...
    def get_vertex(self, id_: str) -> Vertex | None: ...
    def add_vertex(self, vertex: Vertex | None) -> None: ...
    def remove_vertex(self, vertex: Vertex) -> None: ...
    def get_edge(self, vertex: Vertex, another_vertex: Vertex) -> Edge | None: ...
    def add_edge(self, edge: Edge | None) -> None: ...
    def remove_edge(self, edge: Edge) -> None: ...
    def set_all_vertex_marks(self, mark: bool) -> None: ...
    def all_vertices_marked(self) -> bool: ...
    def set_all_edge_marks(self, mark: bool) -> None: ...
    def all_edges_marked(self) -> bool: ...
    def get_neighbours(self, vertex: Vertex) -> List[Vertex]: ...
    def get_edges(self, vertex: Vertex) -> List[Edge]: ...
    @property
    def is_empty(self) -> bool: ...



# pylint: skip-file
__all__: Final[list[str]] = ["Connection", "Client", "Server"]

from abc import ABC, abstractmethod
from typing import Final

class Connection:
    __slots__: Final[tuple[str, str, str]] = ("_socket", "_to_server", "_from_server")
    def __init__(self, server_ip: str, server_port: int) -> None: ...
    def receive(self) -> str | None: ...
    def send(self, message: str) -> None: ...
    def close(self) -> None: ...

class Client(ABC):
    __slots__: Final[tuple[str, str]] = ("_socket_wrapper", "_active")
    def __init__(self, server_ip: str, server_port: int) -> None: ...
    @property
    def is_connected(self) -> bool: ...
    def send(self, message: str) -> None: ...
    def close(self) -> None: ...
    @abstractmethod
    def process_message(self, message: str) -> None: ...

class Server(ABC):
    __slots__: Final[tuple[str, str, str, str]] = (
        "__weakref__",
        "_connection_handler",
        "_message_handlers",
        "_lock",
    )
    def __init__(self, port: int) -> None: ...
    @property
    def is_open(self) -> bool: ...
    def is_connected_to(self, client_ip: str, client_port: int) -> bool: ...
    def send(self, client_ip: str, client_port: int, message: str) -> None: ...
    def send_to_all(self, message: str) -> None: ...
    def close_connection(self, client_ip: str, client_port: int) -> None: ...
    def close(self) -> None: ...
    @abstractmethod
    def process_new_connection(self, client_ip: str, client_port: int) -> None: ...
    @abstractmethod
    def process_message(
        self,
        client_ip: str | None,
        client_port: int,
        message: str,
    ) -> None: ...
    @abstractmethod
    def process_closing_connection(
        self,
        client_ip: str | None,
        client_port: int,
    ) -> None: ...



"""Sortieralgorithmen für `List[ComparableContentT]`."""

from __future__ import annotations

__all__: Final[list[str]] = [
    "bubble_sort",
    "selection_sort",
    "insertion_sort",
    "merge_sort",
    "quick_sort",
]

from typing import TYPE_CHECKING, Final

from nrw.datastructures._list import List

if TYPE_CHECKING:
    from nrw.datastructures._comparable_content import ComparableContentT


def bubble_sort(lst: List[ComparableContentT]) -> List[ComparableContentT]:
    swapped: bool = True
    while swapped:
        swapped = False
        lst.to_first()
        while lst.has_access and lst._current.next_node is not None:
            if lst._current.content > lst._current.next_node.content:
                lst._current.content, lst._current.next_node.content = (
                    lst._current.next_node.content,
                    lst._current.content,
                )
                swapped = True
            lst.next()
    return lst


def selection_sort(lst: List[ComparableContentT]) -> List[ComparableContentT]:
    result: List[ComparableContentT] = List()
    while not lst.is_empty:
        min_element: ComparableContentT = _get_min_element(lst)
        _delete(min_element, lst)
        result.append(min_element)
    return result


def _get_min_element(lst: List[ComparableContentT]) -> ComparableContentT:
    lst.to_first()
    min_element: ComparableContentT | None = lst.content
    assert min_element is not None
    lst.next()
    while lst.has_access:
        assert lst.content is not None
        min_element = min(min_element, lst.content)
        lst.next()
    return min_element


def _delete(element: ComparableContentT, lst: List[ComparableContentT]) -> None:
    lst.to_first()
    while lst.has_access:  # pragma: no branch
        if lst.content is element:
            lst.remove()
            return
        lst.next()


def insertion_sort(lst: List[ComparableContentT]) -> List[ComparableContentT]:
    result: List[ComparableContentT] = List()
    lst.to_first()
    while lst.has_access:
        _insort(result, lst.content)  # type: ignore[misc]
        lst.next()
    return result


def _insort(lst: List[ComparableContentT], element: ComparableContentT) -> None:
    lst.to_first()
    while lst.has_access:
        if element < lst.content:
            lst.insert(element)
            return
        lst.next()
    lst.append(element)


def merge_sort(lst: List[ComparableContentT]) -> List[ComparableContentT]:
    lst.to_first()
    lst.next()
    if not lst.has_access:
        return lst

    left: List[ComparableContentT] = List()
    right: List[ComparableContentT] = List()

    to_the_left: bool = True
    lst.to_first()
    while lst.has_access:
        if to_the_left:
            left.append(lst.content)
            to_the_left = False
        else:
            right.append(lst.content)
            to_the_left = True
        lst.next()

    return _merge(merge_sort(left), merge_sort(right))


def _merge(
    lst1: List[ComparableContentT],
    lst2: List[ComparableContentT],
) -> List[ComparableContentT]:
    if lst1.is_empty:  # pragma: no branch
        return lst2  # pragma: no cover
    if lst2.is_empty:  # pragma: no branch
        return lst1  # pragma: no cover

    result: List[ComparableContentT] = List()

    lst1.to_first()
    lst2.to_first()
    while lst1.has_access and lst2.has_access:
        if lst1.content < lst2.content:  # type: ignore[operator]
            result.append(lst1.content)
            lst1.next()
        else:
            result.append(lst2.content)
            lst2.next()

    while lst1.has_access:
        result.append(lst1.content)
        lst1.next()

    while lst2.has_access:
        result.append(lst2.content)
        lst2.next()

    return result


def quick_sort(lst: List[ComparableContentT]) -> List[ComparableContentT]:
    result: List[ComparableContentT] = List()

    if lst.is_empty:
        return result

    lst.to_first()
    pivot: ComparableContentT | None = lst.content
    assert pivot is not None
    lst.next()

    right: List[ComparableContentT] = List()
    left: List[ComparableContentT] = List()

    while lst.has_access:
        element: ComparableContentT | None = lst.content
        assert element is not None
        if element < pivot:
            left.append(element)
        else:
            right.append(element)
        lst.next()

    result.concat(quick_sort(left))
    result.append(pivot)
    result.concat(quick_sort(right))

    return result



"""Suchalgorithmen für `List[ComparableContentT]`."""

from __future__ import annotations

__all__: Final[list[str]] = [
    "linear_search",
    "depth_first_search",
    "breadth_first_search",
]

from typing import TYPE_CHECKING, Final, TypeVar

from nrw.datastructures._list import List

if TYPE_CHECKING:
    from nrw.datastructures._edge import Edge
    from nrw.datastructures._graph import Graph
    from nrw.datastructures._vertex import Vertex

_T = TypeVar("_T")


def linear_search(lst: List[_T], element: _T) -> int:
    lst.to_first()
    index: int = 0
    while lst.has_access:
        if lst.content == element:
            return index
        index += 1
        lst.next()
    return index


def _depth_first_search_impl(graph: Graph, vertex: Vertex) -> List[Vertex]:
    assert graph.get_vertex(vertex.id) is not None
    result: List[Vertex] = List()

    if vertex.is_marked:
        return result

    result.append(vertex)
    vertex.mark = True

    neighbours: List[Vertex] = graph.get_neighbours(vertex)
    neighbours.to_first()
    while neighbours.has_access:
        current: Vertex | None = neighbours.content
        assert current is not None
        result.concat(_depth_first_search_impl(graph, current))
        neighbours.next()

    return result


def depth_first_search(graph: Graph, vertex: Vertex) -> List[Vertex]:
    if graph.is_empty or graph.get_vertex(vertex.id) is None:
        return List()

    graph.set_all_vertex_marks(False)
    graph.set_all_edge_marks(False)
    return _depth_first_search_impl(graph, vertex)


def breadth_first_search(graph: Graph, vertex: Vertex) -> List[Vertex]:
    result: List[Vertex] = List()

    if graph.is_empty or graph.get_vertex(vertex.id) is None:
        return result

    graph.set_all_vertex_marks(False)
    graph.set_all_edge_marks(False)

    vertex.mark = True
    result.append(vertex)

    result.to_first()
    while result.has_access:
        current: Vertex | None = result.content
        assert current is not None
        neighbours: List[Vertex] = graph.get_neighbours(current)
        neighbours.to_first()
        while neighbours.has_access:
            current_neighbour: Vertex | None = neighbours.content
            assert current_neighbour is not None
            if not current_neighbour.is_marked:
                edge: Edge | None = graph.get_edge(current_neighbour, current)
                assert edge is not None
                edge.mark = True
                current_neighbour.mark = True
                result.append(current_neighbour)
            neighbours.next()
        result.next()

    return result



"""Implementation der generischen Klasse `BinaryTree[_T]`."""

from __future__ import annotations

__all__: Final[list[str]] = ["BinaryTree"]

from typing import Final, Generic, TypeVar

from nrw.datastructures._utils import display_binary_node

_T = TypeVar("_T")


class _BTNode(Generic[_T]):
    """Durch diese innere Klasse kann man dafür sorgen, dass ein leerer Baum `None`
    ist, ein nicht-leerer Baum jedoch immer eine nicht-`None`-Wurzel sowie
    nicht-`None`-Teilbäume, ggf. leere Teilbäume hat.
    """

    __slots__: Final[tuple[str, str, str]] = ("_content", "_left", "_right")
    __hash__ = None  # type: ignore[assignment]

    def __init__(self, content: _T) -> None:
        self._content: _T = content
        self._left: BinaryTree[_T] | None = BinaryTree()
        self._right: BinaryTree[_T] | None = BinaryTree()

    def __repr__(self) -> str:
        return (
            f"{self.__class__.__name__}(content={self._content!r}, "
            f"left_tree={self._left!r}, right_tree={self._right!r})"
        )

    def __str__(self) -> str:
        return display_binary_node(self)


class BinaryTree(Generic[_T]):
    """Mithilfe der generischen Klasse `BinaryTree` können beliebig viele
    Inhaltsobjekte in einem Binaerbaum verwaltet werden. Ein
    Objekt der Klasse stellt entweder einen leeren Baum dar oder verwaltet ein
    Inhaltsobjekt sowie einen linken und einen rechten Teilbaum, die ebenfalls
    Objekte der generischen Klasse `BinaryTree` sind.
    """

    __slots__: Final[tuple[str]] = ("_node",)
    __hash__ = None  # type: ignore[assignment]

    def __init__(
        self,
        content: _T | None = None,
        left_tree: BinaryTree[_T] | None = None,
        right_tree: BinaryTree[_T] | None = None,
    ) -> None:
        """Nach dem Aufruf des Konstruktors ohne Parameter existiert ein leerer
        Binaerbaum.

        Wenn der Parameter `content` gegeben und nicht `None` ist, existiert nach dem
        Aufruf des Konstruktors der Binaerbaum und hat `content` als Inhaltsobjekt
        und zwei leere Teilbäume.
        Falls der Parameter `None` ist, wird ein leerer Binaerbaum erzeugt.


        Wenn der Parameter `content` gegeben und nicht `None` ist
        sowie die Parameter `left_tree` und `right_tree` gegenben sind,
        wird ein Binaerbaum mit `content` als Inhalt und den beiden Teilbäumen
        `left_tree` und `right_tree` erzeugt.
        Sind `left_tree` und `right_tree` `None`, wird der entsprechende
        Teilbaum als leerer Binaerbaum eingefügt. So kann es also nie passieren,
        dass linke oder rechte Teilbäume `None` sind. Wenn der Parameter `content`
        `None` ist, wird ein leerer Binaerbaum erzeugt.
        """
        if content is None:
            self._node: _BTNode[_T] | None = None
        else:
            self._node = _BTNode(content)
            self._node._left = left_tree if left_tree is not None else BinaryTree()
            self._node._right = right_tree if right_tree is not None else BinaryTree()

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}(node={self._node!r})"

    def __str__(self) -> str:
        if self.is_empty:
            return ""
        return str(self._node)

    @property
    def is_empty(self) -> bool:
        """Diese Anfrage liefert das Inhaltsobjekt des Binaerbaums. Wenn der Binaerbaum
        leer ist, wird `None` zurueckgegeben.
        """
        return self._node is None

    @property
    def content(self) -> _T | None:
        """Diese Anfrage liefert das Inhaltsobjekt des Binaerbaums. Wenn der Binaerbaum
        leer ist, wird `None` zurückgegeben.
        """
        return self._node._content if not self.is_empty else None

    @content.setter
    def content(self, new_content: _T | None) -> None:
        if new_content is None:
            return

        if self.is_empty:
            self._node = _BTNode(new_content)
            self._node._left = BinaryTree()
            self._node._right = BinaryTree()
        else:
            self._node._content = new_content

    @property
    def left_tree(self) -> BinaryTree[_T] | None:
        """Diese Anfrage liefert den linken Teilbaum des Binaerbaumes. Wenn der
        Binaerbaum leer ist, wird `None` zurückgegeben.
        """
        return self._node._left if not self.is_empty else None

    @left_tree.setter
    def left_tree(self, new_tree: BinaryTree[_T] | None) -> None:
        if self.is_empty or new_tree is None:
            return
        self._node._left = new_tree

    @property
    def right_tree(self) -> BinaryTree[_T] | None:
        """Diese Anfrage liefert den rechten Teilbaum des Binaerbaumes. Wenn der
        Binaerbaum leer ist, wird `None` zurückgegeben.
        """
        return self._node._right if not self.is_empty else None

    @right_tree.setter
    def right_tree(self, new_tree: BinaryTree[_T] | None) -> None:
        if self.is_empty or new_tree is None:
            return
        self._node._right = new_tree



# pylint: skip-file
__all__: Final[list[str]] = [
    "linear_search",
    "depth_first_search",
    "breadth_first_search",
    "bubble_sort",
    "selection_sort",
    "insertion_sort",
    "merge_sort",
    "quick_sort",
    "preorder",
    "inorder",
    "postorder",
    "levelorder",
    "reverse_inorder",
]

from typing import Final, TypeVar, overload

from nrw.datastructures import (
    BinarySearchTree,
    BinaryTree,
    ComparableContentT,
    Graph,
    List,
    Vertex,
)

_T = TypeVar("_T")

def linear_search(lst: List[_T], element: _T) -> int: ...
def depth_first_search(graph: Graph, vertex: Vertex) -> List[Vertex]: ...
def breadth_first_search(graph: Graph, vertex: Vertex) -> List[Vertex]: ...
def bubble_sort(lst: List[ComparableContentT]) -> List[ComparableContentT]: ...
def selection_sort(lst: List[ComparableContentT]) -> List[ComparableContentT]: ...
def insertion_sort(lst: List[ComparableContentT]) -> List[ComparableContentT]: ...
def merge_sort(lst: List[ComparableContentT]) -> List[ComparableContentT]: ...
def quick_sort(lst: List[ComparableContentT]) -> List[ComparableContentT]: ...
@overload
def preorder(tree: BinaryTree[_T], *, reverse: bool = False) -> List[_T]: ...
@overload
def preorder(
    tree: BinarySearchTree[ComparableContentT],
    *,
    reverse: bool = False,
) -> List[ComparableContentT]: ...
@overload
def inorder(tree: BinaryTree[_T], *, reverse: bool = False) -> List[_T]: ...
@overload
def inorder(
    tree: BinarySearchTree[ComparableContentT],
    *,
    reverse: bool = False,
) -> List[ComparableContentT]: ...
@overload
def reverse_inorder(tree: BinaryTree[_T]) -> List[_T]: ...
@overload
def reverse_inorder(
    tree: BinarySearchTree[ComparableContentT],
) -> List[ComparableContentT]: ...
@overload
def postorder(tree: BinaryTree[_T], *, reverse: bool = False) -> List[_T]: ...
@overload
def postorder(
    tree: BinarySearchTree[ComparableContentT],
    *,
    reverse: bool = False,
) -> List[ComparableContentT]: ...
@overload
def levelorder(tree: BinaryTree[_T], *, reverse: bool = False) -> List[_T]: ...
@overload
def levelorder(
    tree: BinarySearchTree[ComparableContentT],
    *,
    reverse: bool = False,
) -> List[ComparableContentT]: ...



"""Die Datenstrukturen nach den Vorgaben des Landes NRW."""

from __future__ import annotations

__all__: Final[list[str]] = [
    "Stack",
    "Queue",
    "List",
    "BinaryTree",
    "ComparableContent",
    "ComparableContentT",
    "Vertex",
    "Edge",
    "BinarySearchTree",
    "Graph",
]


from typing import Final

from nrw.datastructures._binary_search_tree import BinarySearchTree
from nrw.datastructures._binary_tree import BinaryTree
from nrw.datastructures._comparable_content import ComparableContent, ComparableContentT
from nrw.datastructures._edge import Edge
from nrw.datastructures._graph import Graph
from nrw.datastructures._list import List
from nrw.datastructures._queue import Queue
from nrw.datastructures._stack import Stack
from nrw.datastructures._vertex import Vertex



"""Klasse `QueryResult`."""

from __future__ import annotations

__all__: Final[list[str]] = ["QueryResult"]

from typing import Any, Final


class QueryResult:
    """Ein Objekt der Klasse `QueryResult` stellt die Ergebnistabelle einer
    Datenbankanfrage mit Hilfe
    der Klasse `DatabaseConnector` dar. Objekte dieser Klasse werden nur von der
    Klasse `DatabaseConnector` erstellt.
    Die Klasse verfügt über keinen öffentlichen Konstruktor.
    """

    __slots__: Final[tuple[str, str, str]] = ("_data", "_column_names", "_column_types")

    def __init__(
        self,
        data: list[tuple[Any, ...]],
        column_names: tuple[str, ...],
        column_types: tuple[type | str | None, ...],
    ) -> None:
        """Interner Konstruktor."""
        self._data: list[tuple[Any, ...]] = data
        self._column_names: tuple[str, ...] = column_names
        self._column_types: tuple[type | str | None, ...] = column_types

    @property
    def data(self) -> list[tuple[Any, ...]]:
        """Die Anfrage liefert die Einträge der Ergebnistabelle als eine `list`
        welche wiederum `tuple` enthält. Der erste Index stellt die Zeile und der zweite
        die Spalte dar (d. h. Object[zeile][spalte]).
        """
        return self._data

    @property
    def column_names(self) -> tuple[str, ...]:
        """Die Anfrage liefert die Bezeichner der Spalten der Ergebnistabelle als
        `tuple` vom Typ `str` zurück.
        """
        return self._column_names

    @property
    def column_types(self) -> tuple[type | str | None, ...]:
        """Die Anfrage liefert (wenn möglich) die Typenbezeichnung der Spalten der
        Ergebnistabelle als `tuple` vom jeweiligen Typ zurück.
        """
        return self._column_types

    @property
    def row_count(self) -> int:
        """Die Anfrage liefert die Anzahl der Zeilen der Ergebnistabelle als `int`."""
        return len(self._data) if self._data is not None else 0

    @property
    def column_count(self) -> int:
        """Die Anfrage liefert die Anzahl der Spalten der Ergebnistabelle als `int`."""
        assert all(len(self._data[0]) == len(data) for data in self._data[1:])
        return len(self._data[0])



from __future__ import annotations

__all__: Final[list[str]] = ["Server"]

import socket
import sys
import threading
import weakref
from abc import ABC, abstractmethod
from contextlib import suppress
from typing import TYPE_CHECKING, Final

from nrw.datastructures import List

if sys.version_info >= (3, 12):  # pragma: >=3.12 cover
    from typing import override
else:  # pragma: <3.12 cover
    from typing_extensions import override

if TYPE_CHECKING:
    from io import TextIOWrapper


class _NewConnectionHandler(threading.Thread):
    __slots__: Final[tuple[str, str, str]] = ("_server", "_active", "_server_socket")

    def __init__(
        self,
        port: int,
        server: Server,
    ) -> None:
        super().__init__()
        self._server: weakref.ProxyType[Server] = weakref.proxy(
            server,
            self._on_server_shutdown,
        )
        self._active: bool = False
        self._server_socket: socket.socket | None = None
        with suppress(Exception):
            self._server_socket = socket.socket(
                socket.AF_INET,
                socket.SOCK_STREAM,
            )
            self._server_socket.bind(("127.0.0.1", port))
            self._server_socket.listen()
            self._server_socket.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
            self._active = True
            self.start()

    def _on_server_shutdown(
        self,
        server: Server,  # noqa: ARG002 # pylint: disable=W0613
    ) -> None:
        self.close()

    @override
    def run(self) -> None:
        while self._active:
            with suppress(OSError):
                try:
                    client_socket, (client_ip, client_port) = (
                        self._server_socket.accept()
                    )
                    self._server._add_new_client_message_handler(client_socket)
                    self._server.process_new_connection(client_ip, client_port)
                except ReferenceError:
                    self.close()

    def close(self) -> None:
        if self._active:
            self._active = False
            if self._server_socket is not None:
                with suppress(OSError):
                    self._server_socket.close()


class _ClientSocketWrapper:
    __slots__: Final[tuple[str, str, str]] = (
        "_client_socket",
        "_to_client",
        "_from_client",
    )

    def __init__(self, client_socket: socket.socket | None) -> None:
        try:
            self._client_socket: socket.socket | None = client_socket
            self._to_client: TextIOWrapper | None = self._client_socket.makefile(
                mode="w",
                encoding="utf-8",
            )
            self._from_client: TextIOWrapper | None = self._client_socket.makefile(
                mode="r",
                encoding="utf-8",
            )
        except OSError:
            self._client_socket = None
            self._to_client = None
            self._from_client = None

    def receive(self) -> str | None:
        if self._from_client is not None:
            with suppress(OSError, ValueError):
                received_line: str = self._from_client.readline().strip()
                return received_line if received_line else None
        return None

    def send(self, message: str) -> None:
        if self._to_client is not None:
            with suppress(OSError, ValueError):
                self._to_client.write(f"{message}\n")
                self._to_client.flush()

    @property
    def client_ip(self) -> str | None:
        if self._client_socket is None:
            return None
        try:
            return self._client_socket.getpeername()[0]  # type: ignore[no-any-return]
        except OSError:
            return None

    @property
    def client_port(self) -> int:
        if self._client_socket is None:
            return 0
        try:
            return self._client_socket.getpeername()[1]  # type: ignore[no-any-return]
        except OSError:
            return 0

    def close(self) -> None:
        if self._client_socket is not None:
            with suppress(OSError):
                self._to_client.close()
                self._from_client.close()
                self._client_socket.shutdown(socket.SHUT_RDWR)
                self._client_socket.close()


class _ClientMessageHandler(threading.Thread):
    __slots__: Final[tuple[str, str, str]] = ("_server", "_active", "_socket_wrapper")

    def __init__(self, client_socket: socket.socket | None, server: Server) -> None:
        super().__init__()
        self._server: weakref.ProxyType[Server] = weakref.proxy(
            server,
            self._on_server_shutdown,
        )
        self._active: bool = False
        self._socket_wrapper: _ClientSocketWrapper = _ClientSocketWrapper(client_socket)
        if client_socket is not None:
            self._active = True
            self.start()

    def _on_server_shutdown(
        self,
        server: Server,  # noqa: ARG002 # pylint: disable=W0613
    ) -> None:
        self.close()

    @override
    def run(self) -> None:
        message: str | None = None
        while self._active:
            message = self._socket_wrapper.receive()
            client_ip: str | None = self._socket_wrapper.client_ip
            client_port: int = self._socket_wrapper.client_port
            try:
                if message is not None:
                    self._server.process_message(client_ip, client_port, message)
                else:
                    message_handler: _ClientMessageHandler | None = (
                        self._server._find_client_message_handler(
                            client_ip,
                            client_port,
                        )
                    )
                    if message_handler is not None:
                        message_handler.close()
                        self._server._remove_client_message_handler(message_handler)
                        self._server.process_closing_connection(client_ip, client_port)
            except ReferenceError:
                self.close()

    def send(self, message: str) -> None:
        if self._active:
            self._socket_wrapper.send(message)

    def close(self) -> None:
        if self._active:
            self._active = False
            self._socket_wrapper.close()

    @property
    def client_ip(self) -> str | None:
        return self._socket_wrapper.client_ip

    @property
    def client_port(self) -> int:
        return self._socket_wrapper.client_port


class Server(ABC):
    """Objekte von Unterklassen der abstrakten Klasse `Server` ermöglichen das Anbieten
    von Serverdiensten, so dass Clients Verbindungen zum Server mittels TCP/IP-Protokoll
    aufbauen können. Zur Vereinfachung finden Nachrichtenversand und -empfang
    zeilenweise statt, d. h., beim Senden einer Zeichenkette wird ein Zeilentrenner
    ergänzt und beim Empfang wird dieser entfernt. Verbindungsannahme,
    Nachrichtenempfang und Verbindungsende geschehen nebenläufig. Auf diese Ereignisse
    muss durch Überschreiben der entsprechenden Ereignisbehandlungsmethoden reagiert
    werden. Es findet nur eine rudimentäre Fehlerbehandlung statt, so dass z. B.
    Verbindungsabbrüche nicht zu einem Programmabbruch führen. Einmal unterbrochene oder
    getrennte Verbindungen können nicht reaktiviert werden.
    """

    __slots__: Final[tuple[str, str, str, str]] = (
        "__weakref__",
        "_lock",
        "_message_handlers",
        "_connection_handler",
    )

    def __init__(self, port: int) -> None:
        """Ein Objekt vom Typ `Server` wird erstellt, das über die angegebene Portnummer
        einen Dienst anbietet an. Clients können sich mit dem Server verbinden, so dass
        Daten (Zeichenketten) zu diesen gesendet und von diesen empfangen werden können.
        Kann der Server unter der angegebenen Portnummer keinen Dienst anbieten (z. B.
        weil die Portnummer bereits belegt ist), ist keine Verbindungsaufnahme zum
        Server und kein Datenaustausch möglich.
        """
        self._lock: threading.Lock = threading.Lock()
        self._message_handlers: List[_ClientMessageHandler] = List()
        self._connection_handler: _NewConnectionHandler = _NewConnectionHandler(
            port,
            self,
        )

    def _add_new_client_message_handler(self, client_socket: socket.socket) -> None:
        with self._lock:
            self._message_handlers.append(_ClientMessageHandler(client_socket, self))

    def _remove_client_message_handler(
        self,
        client_message_handler: _ClientMessageHandler,
    ) -> None:
        with self._lock:
            self._message_handlers.to_first()
            while self._message_handlers.has_access:
                if client_message_handler is self._message_handlers.content:
                    self._message_handlers.remove()
                    return
                self._message_handlers.next()

    def _find_client_message_handler(
        self,
        client_ip: str | None,
        client_port: int,
    ) -> _ClientMessageHandler | None:
        with self._lock:
            self._message_handlers.to_first()
            while self._message_handlers.has_access:
                current_message_handler: _ClientMessageHandler | None = (
                    self._message_handlers.content
                )
                assert current_message_handler is not None
                if (
                    current_message_handler.client_ip == client_ip
                    and current_message_handler.client_port == client_port
                ):
                    return current_message_handler
                self._message_handlers.next()
            return None

    @property
    def is_open(self) -> bool:
        """Die Anfrage liefert den Wert `True`, wenn der `Server` auf Port `port` einen
        Dienst anbietet. Ansonsten liefert die Methode den Wert `False`.
        """
        return self._connection_handler._active

    def is_connected_to(self, client_ip: str, client_port: int) -> bool:
        """Die Anfrage liefert den Wert `True`, wenn der Server mit dem durch
        `client_ip` und `client_port` spezifizierten Client aktuell verbunden ist.
        Ansonsten liefert die Methode den Wert `False`.
        """
        return self._find_client_message_handler(client_ip, client_port) is not None

    def send(self, client_ip: str, client_port: int, message: str) -> None:
        """Die Nachricht `message` wird - um einen Zeilentrenner erweitert - an den
        durch `client_ip` und `client_port` spezifizierten Client gesendet. Schlägt der
        Versand fehl, geschieht nichts.
        """
        message_handler: _ClientMessageHandler | None = (
            self._find_client_message_handler(client_ip, client_port)
        )
        if message_handler is not None:
            message_handler.send(message)

    def send_to_all(self, message: str) -> None:
        """Die Nachricht `message` wird - um einen Zeilentrenner erweitert - an alle mit
        dem Server verbundenen Clients gesendet. Schlägt der Versand an einen Client
        fehl, wird dieser Client übersprungen.
        """
        with self._lock:
            self._message_handlers.to_first()
            while self._message_handlers.has_access:
                self._message_handlers.content.send(message)
                self._message_handlers.next()

    def close_connection(self, client_ip: str, client_port: int) -> None:
        """Die Verbindung des Servers zu dem durch `client_ip` und `client_port`
        spezifizierten Client wird getrennt. Zuvor wird die Methode
        `process_closing_connection` mit IP-Adresse und Port des jeweiligen Clients
        aufgerufen. Ist der Server nicht mit dem in der Parameterliste spezifizierten
        Client verbunden, geschieht nichts.
        """
        message_handler: _ClientMessageHandler | None = (
            self._find_client_message_handler(client_ip, client_port)
        )
        if message_handler is not None:
            self.process_closing_connection(client_ip, client_port)
            message_handler.close()
            self._remove_client_message_handler(message_handler)

    def close(self) -> None:
        """Alle bestehenden Verbindungen zu Clients werden getrennt und der Server kann
        nicht mehr verwendet werden. Ist der Server bereits vor Aufruf der Methode in
        diesem Zustand, geschieht nichts.
        """
        self._connection_handler.close()
        with self._lock:
            self._message_handlers.to_first()
            while self._message_handlers.has_access:
                message_handler: _ClientMessageHandler | None = (
                    self._message_handlers.content
                )
                assert message_handler is not None
                self.process_closing_connection(
                    message_handler.client_ip,
                    message_handler.client_port,
                )
                self._message_handlers.remove()
                self._message_handlers.next()

    @abstractmethod
    def process_new_connection(
        self,
        client_ip: str,
        client_port: int,
    ) -> None:
        """Diese Ereignisbehandlungsmethode wird aufgerufen, wenn sich ein Client mit
        IP-Adresse `client_ip` und Portnummer `client_port` mit dem Server verbunden
        hat. Die Methode ist abstrakt und muss in einer Unterklasse der Klasse `Server`
        überschrieben werden, so dass auf den Neuaufbau der Verbindung reagiert wird.
        Der Aufruf der Methode erfolgt nicht synchronisiert.
        """

    @abstractmethod
    def process_message(
        self,
        client_ip: str | None,
        client_port: int,
        message: str,
    ) -> None:
        """Diese Ereignisbehandlungsmethode wird aufgerufen, wenn der Server die
        Nachricht `message` von dem durch `client_ip` und `client_port` spezifizierten
        Client empfangen hat. Der vom Client hinzugefügte Zeilentrenner wurde zuvor
        entfernt. Die Methode ist abstrakt und muss in einer Unterklasse der Klasse
        `Server` überschrieben werden, so dass auf den Empfang der Nachricht reagiert
        wird. Der Aufruf der Methode erfolgt nicht synchronisiert.
        """

    @abstractmethod
    def process_closing_connection(
        self,
        client_ip: str | None,
        client_port: int,
    ) -> None:
        """Sofern der Server die Verbindung zu dem durch `client_ip` und `client_port`
        spezifizierten Client trennt, wird diese Ereignisbehandlungsmethode aufgerufen,
        unmittelbar bevor die Verbindungstrennung tatsächlich erfolgt.
        Wird die Verbindung unvermittelt unterbrochen oder hat der in der Parameterliste
        spezifizierte Client die Verbindung zum Server unvermittelt getrennt, erfolgt
        der Methodenaufruf nach der Unterbrechung / Trennung der Verbindung.
        Die Methode ist abstrakt und muss in einer Unterklasse der Klasse Server
        überschrieben werden, so dass auf das Ende der Verbindung zum angegebenen Client
        reagiert wird. Der Aufruf der Methode erfolgt nicht synchronisiert.
        """



# pylint: skip-file
__all__: Final[list[str]] = ["QueryResult"]

from typing import Any, Final

class QueryResult:
    __slots__: Final[tuple[str, str, str]] = ("_data", "_column_names", "_column_types")
    def __init__(
        self,
        data: list[tuple[Any, ...]],
        column_names: tuple[str, ...],
        column_types: tuple[type | str | None, ...],
    ) -> None: ...
    @property
    def data(self) -> list[tuple[Any, ...]]: ...
    @property
    def column_names(self) -> tuple[str, ...]: ...
    @property
    def column_types(self) -> tuple[type | str | None, ...]: ...
    @property
    def row_count(self) -> int: ...
    @property
    def column_count(self) -> int: ...



"""Klasse `DatabaseConnector`."""

from __future__ import annotations

__all__: Final[list[str]] = ["DatabaseConnector"]

from typing import TYPE_CHECKING, Final

import mysql.connector
from mysql.connector import FieldType

from nrw.database._query_result import QueryResult

if TYPE_CHECKING:
    from mysql.connector.abstracts import MySQLConnectionAbstract
    from mysql.connector.pooling import PooledMySQLConnection


class DatabaseConnector:
    """Ein Objekt der Klasse `DatabaseConnector` ermöglicht die Abfrage und Manipulation
    einer `MySQL`-Datenbank.
    Beim Erzeugen des Objekts wird eine Datenbankverbindung aufgebaut, so dass
    anschließend SQL-Anweisungen an diese Datenbank gerichtet werden können.
    """

    # pylint: disable=W0718, R0913

    __slots__: Final[tuple[str, str, str]] = (
        "_connection",
        "_current_query_result",
        "_message",
    )

    def __init__(
        self,
        ip: str,
        port: int,
        database: str,
        username: str,
        password: str,
    ) -> None:
        """Ein Objekt vom Typ `DatabaseConnector` wird erstellt, und eine Verbindung zur
        Datenbank wird aufgebaut. Mit den Parametern `ip` und `port` werden die
        IP-Adresse und die Port-Nummer übergeben, unter denen die Datenbank mit Namen
        `database` zu erreichen ist.
        Mit den Parametern `username` und `password` werden Benutzername und Passwort
        für die Datenbank übergeben.
        """
        self._current_query_result: QueryResult | None = None
        self._message: str | None = None

        try:
            self._connection: PooledMySQLConnection | MySQLConnectionAbstract | None = (
                mysql.connector.connect(
                    user=username,
                    password=password,
                    host=ip,
                    port=port,
                    database=database,
                    autocommit=True,
                )
            )
        except Exception as exception:
            self._connection = None
            self._message = str(exception)

    def execute_statement(self, sql_statement: str) -> None:
        """Der Auftrag schickt den im Parameter `sql_statement` enthaltenen SQL-Befehl
        an die Datenbank ab.
        Handelt es sich bei `sql_statement` um einen SQL-Befehl, der eine Ergebnismenge
        liefert, so kann dieses Ergebnis anschließend mit dem Property
        `current_query_result` abgerufen werden.
        """
        self._current_query_result = None
        self._message = None

        if self._connection is None:
            self._message = "No connection"
            return

        try:
            with self._connection.cursor(dictionary=False) as cursor:
                cursor.execute(sql_statement)
                if data := cursor.fetchall():
                    assert cursor.description is not None, "No description"
                    column_names: tuple[str, ...] = tuple(
                        column[0] for column in cursor.description
                    )
                    colum_types: tuple[str | None, ...] = tuple(
                        FieldType.get_info(column[1]) for column in cursor.description  # type: ignore[misc]
                    )
                    self._current_query_result = QueryResult(
                        data,  # type: ignore[arg-type]
                        column_names,
                        colum_types,
                    )
        except Exception as exception:
            self._message = str(exception)

    @property
    def current_query_result(self) -> QueryResult | None:
        """Die Anfrage liefert das Ergebnis des letzten mit der Methode
        `execute_statement` an die Datenbank geschickten SQL-Befehls als Objekt vom Typ
        `QueryResult` zurück.
        Wurde bisher kein SQL-Befehl abgeschickt oder ergab der letzte Aufruf von
        `execute_statement` keine Ergebnismenge (z.B. bei einem INSERT-Befehl oder einem
        Syntaxfehler), so wird `None` geliefert.
        """
        return self._current_query_result

    @property
    def error_message(self) -> str | None:
        """Die Anfrage liefert `None` oder eine Fehlermeldung, die sich jeweils auf die
        letzte zuvor ausgefuehrte Datenbankoperation bezieht.
        """
        return self._message

    def close(self) -> None:
        """Die Datenbankverbindung wird geschlossen."""
        if self._connection is None:
            self._message = "No connection"
            return

        try:
            self._connection.close()
        except Exception as exception:
            self._message = str(exception)



"""Implementation der generischen Klasse `List[_T]`."""

from __future__ import annotations

__all__: Final[list[str]] = ["List"]

from io import StringIO
from typing import Final, Generic, TypeVar

_T = TypeVar("_T")


class _ListNode(Generic[_T]):
    __slots__: Final[tuple[str, str]] = ("_content", "_next_node")
    __hash__ = None  # type: ignore[assignment]

    def __init__(self, content: _T) -> None:
        """Ein neues Objekt vom Typ `_ListNode[_T]` wird erschaffen.
        Der Inhalt wird per Parameter gesetzt. Der Verweis ist leer.
        """
        self._content: _T = content
        self._next_node: _ListNode[_T] | None = None

    def __repr__(self) -> str:
        return (
            f"{self.__class__.__name__}(content={self._content!r}, "
            f"next_node={self._next_node!r})"
        )

    @property
    def content(self) -> _T:
        """Liefert das Inhaltsobjekt des Knotens."""
        return self._content

    @content.setter
    def content(self, new_content: _T) -> None:
        self._content = new_content

    @property
    def next_node(self) -> _ListNode[_T] | None:
        """Liefert das nächste Element des aktuellen Knotens."""
        return self._next_node

    @next_node.setter
    def next_node(self, new_next_node: _ListNode[_T] | None) -> None:
        self._next_node = new_next_node


class List(Generic[_T]):
    """Objekt der generischen Klasse `List` verwalten beliebig viele linear
    angeordnete Objekte. Auf höchstens ein Listenobjekt,
    aktuellesObjekt genannt, kann jeweils zugegriffen werden.
    Wenn eine Liste leer ist, vollständig durchlaufen wurde oder das aktuelle
    Objekt am Ende der Liste gelöscht wurde, gibt es kein aktuelles
    Objekt.

    Das erste oder das letzte Objekt einer Liste können durch einen Auftrag zum
    aktuellen Objekt gemacht werden. Außerdem kann das dem aktuellen Objekt
    folgende Listenobjekt zum neuen aktuellen Objekt werden.
    Das aktuelle Objekt kann gelesen, verändert oder gelöscht werden. Außerdem
    kann vor dem aktuellen Objekt ein Listenobjekt eingefügt werden.
    """

    __slots__: Final[tuple[str, str, str]] = ("_first", "_last", "_current")
    __hash__ = None  # type: ignore[assignment]

    def __init__(self) -> None:
        """Eine leere Liste wird erzeugt."""
        self._first: _ListNode[_T] | None = None
        self._last: _ListNode[_T] | None = None
        self._current: _ListNode[_T] | None = None

    def __repr__(self) -> str:
        return (
            f"{self.__class__.__name__}(first={self._first!r}, last={self._last!r}, "
            f"current={self._current!r})"
        )

    def __str__(self) -> str:
        if self.is_empty:
            return f"{self.__class__.__name__}()"

        with StringIO() as buffer:
            buffer.write(f"{self.__class__.__name__}(")
            temp: _ListNode[_T] | None = self._first
            while temp is not self._last:
                buffer.write(f"{temp.content} -> ")
                temp = temp.next_node
            buffer.write(f"{temp.content})")
            return buffer.getvalue()

    @property
    def is_empty(self) -> bool:
        """Die Anfrage liefert den Wert `True`, wenn die Liste keine Objekte enthält,
        sonst liefert sie den Wert `False`.
        """
        return self._first is None

    @property
    def has_access(self) -> bool:
        """Die Anfrage liefert den Wert `True`, wenn es ein aktuelles Objekt gibt,
        sonst liefert sie den Wert `False`.
        """
        return self._current is not None

    def next(self) -> None:
        """Falls die Liste nicht leer ist, es ein aktuelles Objekt gibt und dieses
        nicht das letzte Objekt der Liste ist, wird das dem aktuellen Objekt in
        der Liste folgende Objekt zum aktuellen Objekt, andernfalls gibt es nach
        Ausführung des Auftrags kein aktuelles Objekt, d. h. `has_access` liefert
        den Wert `False`.
        """
        if not self.has_access:
            return
        self._current = self._current.next_node

    def to_first(self) -> None:
        """Falls die Liste nicht leer ist, wird das erste Objekt der Liste aktuelles
        Objekt. Ist die Liste leer, geschieht nichts.
        """
        if self.is_empty:
            return
        self._current = self._first

    def to_last(self) -> None:
        """Falls die Liste nicht leer ist, wird das letzte Objekt der Liste
        aktuelles Objekt. Ist die Liste leer, geschieht nichts.
        """
        if self.is_empty:
            return
        self._current = self._last

    @property
    def content(self) -> _T | None:
        """Falls es ein aktuelles Objekt gibt (`has_access is True`), wird das
        aktuelle Objekt zurückgegeben, andernfalls (`has_access is False`) gibt
        die Anfrage den Wert `None` zurück.
        """
        return self._current.content if self.has_access else None

    @content.setter
    def content(self, new_content: _T | None) -> None:
        if new_content is None or not self.has_access:
            return
        self._current.content = new_content

    def insert(self, content: _T | None) -> None:
        """Falls es ein aktuelles Objekt gibt (`has_access is True`), wird ein neues
        Objekt vor dem aktuellen Objekt in die Liste eingefügt. Das aktuelle
        Objekt bleibt unverändert.

        Wenn die Liste leer ist, wird `content` in die Liste eingefügt und es
        gibt weiterhin kein aktuelles Objekt (`has_access is False`).

        Falls es kein aktuelles Objekt gibt (`has_access is False`) und die Liste
        nicht leer ist oder `content` `None` ist, geschieht nichts.
        """
        if content is None:
            return

        new_node: _ListNode[_T] = _ListNode(content)
        if self.has_access:
            if self._current is not self._first:
                previous: _ListNode[_T] | None = self._get_previous(
                    self._current,
                )
                assert previous is not None
                new_node.next_node, previous.next_node = previous.next_node, new_node
            else:
                new_node.next_node, self._first = self._first, new_node
        elif self.is_empty:
            self._first = self._last = new_node

    def append(self, content: _T | None) -> None:
        """Falls `content` `None` ist, geschieht nichts.

        Ansonsten wird ein neues Objekt `content` am Ende der Liste eingefügt.
        Das aktuelle Objekt bleibt unverändert.

        Wenn die Liste leer ist, wird das Objekt `content` in die Liste eingefügt
        und es gibt weiterhin kein aktuelles Objekt (`has_access is False`).
        """
        if content is None:
            return

        new_node: _ListNode[_T] = _ListNode(content)
        if self.is_empty:
            self._first = self._last = new_node
        else:
            self._last.next_node = self._last = new_node

    def concat(self, other_list: List[_T] | None) -> None:
        """Falls es sich bei der Liste und `other_list` um dasselbe Objekt handelt,
        `other_list` `None` oder eine leere Liste ist, geschieht nichts.

        Ansonsten wird die Liste `other_list` an die aktuelle Liste angehängt.
        Anschliessend wird `other_list` eine leere Liste. Das aktuelle Objekt bleibt
        unverändert. Insbesondere bleibt `has_access` identisch.
        """
        if other_list is self or other_list is None or other_list.is_empty:
            return

        if self.is_empty:
            self._first, self._last = other_list._first, other_list._last
        else:
            self._last.next_node, self._last = other_list._first, other_list._last

        other_list._first = other_list._last = other_list._current = None

    def remove(self) -> None:
        """Wenn die Liste leer ist oder es kein aktuelles Objekt gibt (`has_access
        is False`), geschieht nichts.

        Falls es ein aktuelles Objekt gibt (`has_access is True`), wird das
        aktuelle Objekt gelöscht und das Objekt hinter dem gelöschten Objekt
        wird zum aktuellen Objekt.

        Wird das Objekt, das am Ende der Liste steht, gelöscht, gibt es kein
        aktuelles Objekt mehr.
        """
        if not self.has_access or self.is_empty:
            return

        if self._current is self._first:
            self._first = self._first.next_node
        else:
            previous: _ListNode[_T] | None = self._get_previous(self._current)
            assert previous is not None
            if self._current is self._last:
                self._last = previous
            previous.next_node = self._current.next_node

        self._current = self._current.next_node

        if self.is_empty:
            self._last = None

    def _get_previous(self, node: _ListNode[_T] | None) -> _ListNode[_T] | None:
        """Liefert den Vorgängerknoten des Knotens `node`. Ist die Liste leer, `node
        is None`, `node` nicht in der Liste oder `node` der erste Knoten der Liste,
        wird `None` zurückgegeben.
        """
        if node is None or node == self._first or self.is_empty:
            return None

        temp: _ListNode[_T] | None = self._first
        while temp is not None and temp.next_node is not node:
            temp = temp.next_node
        return temp



from __future__ import annotations

__all__: Final[list[str]] = ["display_binary_node"]

from typing import TYPE_CHECKING, Final, TypeVar

if TYPE_CHECKING:
    from nrw.datastructures._binary_search_tree import _BSTNode
    from nrw.datastructures._binary_tree import _BTNode
    from nrw.datastructures._comparable_content import ComparableContentT

_T = TypeVar("_T")


def display_binary_node(node: _BTNode[_T] | _BSTNode[ComparableContentT]) -> str:
    lines, *_ = _display_binary_node_impl(node)
    return "\n".join(lines)


def _display_binary_node_impl(  # pylint: disable=R0914
    node: _BTNode[_T] | _BSTNode[ComparableContentT],
) -> tuple[list[str], int, int, int]:
    """Based on the implementation by J.V.

    https://stackoverflow.com/questions/34012886/print-binary-tree-level-by-level-in-python/54074933#54074933
    """
    if (node._right is None or node._right.is_empty) and (
        node._left is None or node._left.is_empty
    ):
        line: str = str(node._content)
        width: int = len(line)
        height: int = 1
        middle: int = width // 2
        return [line], width, height, middle

    content: str
    content_len: int
    first_line: str
    second_line: str
    shifted_lines: list[str]
    if (node._right is None or node._right.is_empty) and not node._left.is_empty:
        assert node._left._node is not None
        lines, n, p, x = _display_binary_node_impl(node._left._node)
        content = str(node._content)
        content_len = len(content)
        first_line = (x + 1) * " " + (n - x - 1) * "_" + content
        second_line = x * " " + "/" + (n - x - 1 + content_len) * " "
        shifted_lines = [line + content_len * " " for line in lines]
        return (
            [first_line, second_line, *shifted_lines],
            n + content_len,
            p + 2,
            n + content_len // 2,
        )

    if (node._left is None or node._left.is_empty) and not node._right.is_empty:
        assert node._right._node is not None
        lines, n, p, x = _display_binary_node_impl(node._right._node)
        content = str(node._content)
        content_len = len(content)
        first_line = content + x * "_" + (n - x) * " "
        second_line = (content_len + x) * " " + "\\" + (n - x - 1) * " "
        shifted_lines = [content_len * " " + line for line in lines]
        return (
            [first_line, second_line, *shifted_lines],
            n + content_len,
            p + 2,
            content_len // 2,
        )

    assert node._left._node is not None
    assert node._right._node is not None
    left, n, p, x = _display_binary_node_impl(node._left._node)
    right, m, q, y = _display_binary_node_impl(node._right._node)
    content = str(node._content)
    content_len = len(content)
    first_line = (x + 1) * " " + (n - x - 1) * "_" + content + y * "_" + (m - y) * " "
    second_line = (
        x * " " + "/" + (n - x - 1 + content_len + y) * " " + "\\" + (m - y - 1) * " "
    )
    if p < q:  # pragma: no cover
        left += [n * " "] * (q - p)
    elif q < p:  # pragma: no cover
        right += [m * " "] * (p - q)
    zipped_lines = zip(left, right)
    lines = [first_line, second_line] + [
        a + content_len * " " + b for a, b in zipped_lines
    ]
    return lines, n + m + content_len, max(p, q) + 2, n + content_len // 2



"""Die Netzklassen nach den Vorgaben des Landes NRW."""

from __future__ import annotations

__all__: Final[list[str]] = ["Connection", "Client", "Server"]

from typing import Final

from nrw.network._client import Client
from nrw.network._connection import Connection
from nrw.network._server import Server



"""Implementation der generischen Klasse `Stack[_T]`."""

from __future__ import annotations

__all__: Final[list[str]] = ["Stack"]

from io import StringIO
from typing import Final, Generic, TypeVar

_T = TypeVar("_T")


class _StackNode(Generic[_T]):
    __slots__: Final[tuple[str, str]] = ("_content", "_next_node")
    __hash__ = None  # type: ignore[assignment]

    def __init__(self, content: _T) -> None:
        """Ein neues Objekt vom Typ `_StackNode[_T]` wird erschaffen.
        Der Inhalt wird per Parameter gesetzt. Der Verweis ist leer.
        """
        self._content: _T = content
        self._next_node: _StackNode[_T] | None = None

    def __repr__(self) -> str:
        return (
            f"{self.__class__.__name__}(content={self._content!r}, "
            f"next_node={self._next_node!r})"
        )

    @property
    def content(self) -> _T:
        """Liefert das Inhaltsobjekt des Knotens."""
        return self._content

    @property
    def next_node(self) -> _StackNode[_T] | None:
        """Liefert das nächste Element des aktuellen Knotens."""
        return self._next_node

    @next_node.setter
    def next_node(self, new_next_node: _StackNode[_T] | None) -> None:
        self._next_node = new_next_node


class Stack(Generic[_T]):
    """Objekte der generischen Klasse `Stack` (Keller, Stapel) verwalten beliebige
    Objekte nach dem Last-In-First-Out-Prinzip, d. h., das
    zuletzt abgelegte Objekt wird als erstes wieder entnommen. Alle Methoden
    haben eine konstante Laufzeit, unabhängig von der Anzahl der verwalteten
    Objekte.
    """

    __slots__: Final[tuple[str]] = ("_head",)
    __hash__ = None  # type: ignore[assignment]

    def __init__(self) -> None:
        """Ein leerer Stapel wird erzeugt."""
        self._head: _StackNode[_T] | None = None

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}(head={self._head!r})"

    def __str__(self) -> str:
        if self.is_empty:
            return f"{self.__class__.__name__}()"

        with StringIO() as buffer:
            buffer.write(f"{self.__class__.__name__}(")
            temp: _StackNode[_T] | None = self._head
            while temp is not None and temp.next_node is not None:
                buffer.write(f"{temp.content} -> ")
                temp = temp.next_node
            buffer.write(f"{temp.content})")
            return buffer.getvalue()

    @property
    def is_empty(self) -> bool:
        """Die Anfrage liefert den Wert `True`, wenn der Stapel keine Objekte enthält,
        sonst liefert sie den Wert `False`.
        """
        return self._head is None

    def push(self, content: _T) -> None:
        """Das Objekt `content` wird oben auf den Stapel gelegt.
        Falls `content` `None` ist, bleibt der Stapel unverändert.
        """
        if content is None:
            return

        new_node: _StackNode[_T] = _StackNode(content)
        new_node.next_node, self._head = self._head, new_node

    def pop(self) -> None:
        """Das zuletzt eingefügte Objekt wird von dem Stapel entfernt.
        Falls der Stapel leer ist, bleibt er unverändert.
        """
        if self.is_empty:
            return
        self._head = self._head.next_node

    @property
    def top(self) -> _T | None:
        """Die Anfrage liefert das oberste Stapelobjekt. Der Stapel bleibt unverändert.
        Falls der Stapel leer ist, wird `None` zurückgegeben.
        """
        return self._head.content if not self.is_empty else None



"""Implementation der Klasse `Vertex`."""

from __future__ import annotations

__all__: Final[list[str]] = ["Vertex"]

from typing import Final


class Vertex:
    """Die Klasse Vertex stellt einen einzelnen Knoten eines Graphen dar. Jedes Objekt
    dieser Klasse verfügt über eine im Graphen eindeutige ID als String und kann diese
    ID zurückliefern.
    Darüber hinaus kann eine Markierung gesetzt und abgefragt werden.
    """

    __slots__: Final[tuple[str, str]] = ("_id", "_mark")
    __hash__ = None  # type: ignore[assignment]

    def __init__(self, id_: str) -> None:
        """Ein neues Objekt vom Typ `Vertex` wird erstellt.
        Seine Markierung hat den Wert `False`.
        """
        self._id: str = id_
        self._mark: bool = False

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}(id={self._id!r}, mark={self._mark!r})"

    def __str__(self) -> str:
        return self._id

    @property
    def id(self) -> str:
        """Die Anfrage liefert die ID des Knotens als String."""
        return self._id

    @property
    def mark(self) -> bool:
        """Die Anfrage liefert `True`,
        wenn die Markierung des Knotens den Wert `True` hat, ansonsten `False`.
        """
        return self._mark

    @mark.setter
    def mark(self, new_mark: bool) -> None:
        self._mark = new_mark

    @property
    def is_marked(self) -> bool:
        """Die Anfrage liefert `True`,
        wenn die Markierung des Knotens den Wert `True` hat, ansonsten `False`.
        """
        return self._mark



"""Nützliche Algorithmen für die Datenstrukturen."""

from __future__ import annotations

__all__: Final[list[str]] = [
    "linear_search",
    "depth_first_search",
    "breadth_first_search",
    "bubble_sort",
    "selection_sort",
    "insertion_sort",
    "merge_sort",
    "quick_sort",
    "preorder",
    "inorder",
    "postorder",
    "levelorder",
    "reverse_inorder",
]

from typing import Final

from nrw.algorithms._searching import (
    breadth_first_search,
    depth_first_search,
    linear_search,
)
from nrw.algorithms._sorting import (
    bubble_sort,
    insertion_sort,
    merge_sort,
    quick_sort,
    selection_sort,
)
from nrw.algorithms._traversal import (
    inorder,
    levelorder,
    postorder,
    preorder,
    reverse_inorder,
)



"""Implementation der generischen Klasse `BinarySearchTree[ComparableContentT]`."""

from __future__ import annotations

__all__: Final[list[str]] = ["BinarySearchTree"]

from typing import Final, Generic

from nrw.datastructures._comparable_content import ComparableContentT
from nrw.datastructures._utils import display_binary_node


class _BSTNode(Generic[ComparableContentT]):
    """Durch diese innere Klasse kann man dafür sorgen, dass ein leerer Baum
    `None` ist, ein nicht-leerer Baum jedoch immer eine nicht-`None`-Wurzel sowie
    nicht-`None`-Teilbäume hat.
    """

    __slots__: Final[tuple[str, str, str]] = ("_content", "_left", "_right")
    __hash__ = None  # type: ignore[assignment]

    def __init__(self, content: ComparableContentT) -> None:
        self._content: ComparableContentT = content
        self._left: BinarySearchTree[ComparableContentT] = BinarySearchTree()
        self._right: BinarySearchTree[ComparableContentT] = BinarySearchTree()

    def __repr__(self) -> str:
        return (
            f"{self.__class__.__name__}(content={self._content!r}, "
            f"left_tree={self._left!r}, right_tree={self._right!r})"
        )

    def __str__(self) -> str:
        return display_binary_node(self)


class BinarySearchTree(Generic[ComparableContentT]):
    """Mithilfe der generischen Klasse `BinarySearchTree` können beliebig viele
    Objekte in einem Binaerbaum (binaerer Suchbaum) entsprechend einer
    Ordnungsrelation verwaltet werden.

    Ein Objekt der Klasse stellt entweder einen leeren binären Suchbaum dar oder
    verwaltet ein Inhaltsobjekt sowie einen linken und einen rechten Teilbaum,
    die ebenfalls Objekte der Klasse `BinarySearchTree` sind.

    Die Klasse der Objekte, die in dem Suchbaum verwaltet werden sollen, muss
    das Protocol `ComparableContent` implementieren. Dabei muss durch
    Ueberschreiben der drei Vergleichsmethoden `__lt__`, `__eq__`, `__gt__` (s.
    Dokumentation des Protocols) eine eindeutige Ordnungsrelation festgelegt
    sein.

    Alle Objekte im linken Teilbaum sind kleiner als das Inhaltsobjekt des
    binären Suchbaums. Alle Objekte im rechten Teilbaum sind grösser als das
    Inhaltsobjekt des binären Suchbaums. Diese Bedingung gilt (rekursiv) auch in
    beiden Teilbäumen.

    Hinweis: In dieser Version wird die Klasse `BinaryTree` nicht benutzt.
    """

    __slots__: Final[tuple[str]] = ("_node",)
    __hash__ = None  # type: ignore[assignment]

    def __init__(self) -> None:
        """Der Konstruktor erzeugt einen leeren Suchbaum."""
        self._node: _BSTNode[ComparableContentT] | None = None

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}(node={self._node!r})"

    def __str__(self) -> str:
        return str(self._node) if not self.is_empty else ""

    @property
    def is_empty(self) -> bool:
        """Diese Anfrage liefert den Wahrheitswert `True`, wenn der Suchbaum leer ist,
        sonst liefert sie den Wert `False`.
        """
        return self._node is None

    @property
    def content(self) -> ComparableContentT | None:
        """Diese Anfrage liefert das Inhaltsobjekt des Suchbaumes. Wenn der Suchbaum
        leer ist, wird `None` zurückgegeben.
        """
        return self._node._content if not self.is_empty else None

    @property
    def left_tree(self) -> BinarySearchTree[ComparableContentT] | None:
        """Diese Anfrage liefert den linken Teilbaum des binären Suchbaumes.

        Wenn er leer ist, wird `None` zurückgegeben.
        """
        return self._node._left if not self.is_empty else None

    @property
    def right_tree(self) -> BinarySearchTree[ComparableContentT] | None:
        """Diese Anfrage liefert den rechten Teilbaum des binären Suchbaumes.

        Wenn er leer ist, wird `None` zurückgegeben.
        """
        return self._node._right if not self.is_empty else None

    def insert(self, content: ComparableContentT | None) -> None:
        """Falls der Parameter `None` ist, geschieht nichts.

        Falls ein bezüglich des verwendeten Vergleichs `==` mit
        `content` übereinstimmendes Objekt im geordneten binaeren Suchbau
        enthalten ist, passiert nichts.

        Achtung: hier wird davon ausgegangen, dass `==` genau dann `True`
        liefert, wenn `<` und `>` `False` liefern.

        Andernfalls (`<` order `>`) wird das Objekt `content` entsprechend
        der vorgegebenen Ordnungsrelation in den `BinarySearchTree` eingeordnet.
        """
        if content is None:
            return

        if self.is_empty:
            self._node = _BSTNode(content)
        elif content < self._node._content:
            self._node._left.insert(content)
        elif content > self._node._content:
            self._node._right.insert(content)

    def remove(self, content: ComparableContentT | None) -> None:
        """Falls ein bezüglich des verwendeten Vergleichs mit
        `content` übereinstimmendes Objekt im binaeren Suchbaum enthalten
        ist, wird dieses entfernt. Falls der Parameter `None` ist, ändert sich
        nichts.
        """
        if self.is_empty or content is None:
            return

        if content < self._node._content:
            self._node._left.remove(content)
        elif content > self._node._content:
            self._node._right.remove(content)
        elif self._node._left.is_empty and self._node._right.is_empty:
            self._node = None
        elif self._node._left.is_empty and not self._node._right.is_empty:
            self._node = self._node_of_right_successor
        elif not self._node._left.is_empty and self._node._right.is_empty:
            self._node = self._node_of_left_successor
        elif self._node_of_right_successor._left.is_empty:
            self._node._content, self._node._right = (
                self._node_of_right_successor._content,
                self._node_of_right_successor._right,
            )
        else:
            previous: BinarySearchTree[ComparableContentT] = (
                self._node._right._ancestor_of_small_right()
            )
            smallest: BinarySearchTree[ComparableContentT] | None = previous._node._left
            self._node._content = smallest._node._content
            previous.remove(smallest._node._content)

    def search(self, content: ComparableContentT | None) -> ComparableContentT | None:
        """Falls ein bezüglich des verwendeten Vergleichs `==` mit
        `content` übereinstimmendes Objekt im binaeren Suchbaum enthalten ist,
        liefert die Anfrage dieses, ansonsten wird `None` zurückgegeben.

        Falls der Parameter `None` ist, wird `None` zurückgegeben.
        """
        if self.is_empty or content is None:
            return None

        if content < self._node._content:
            return self.left_tree.search(content)
        if content > self._node._content:
            return self.right_tree.search(content)
        if content == self._node._content:
            return self._node._content
        return None  # pragma: no cover

    def _ancestor_of_small_right(self) -> BinarySearchTree[ComparableContentT]:
        """Die Methode liefert denjenigen Baum, dessen linker Nachfolger keinen linken
        Nachfolger mehr hat. Es ist also später möglich, in einem Baum im
        rechten Nachfolger den Vorgänger des linkesten Nachfolgers zu finden.
        """
        if self._node_of_left_successor._left.is_empty:
            return self
        return self._node._left._ancestor_of_small_right()

    @property
    def _node_of_left_successor(self) -> _BSTNode[ComparableContentT] | None:
        return self._node._left._node

    @property
    def _node_of_right_successor(self) -> _BSTNode[ComparableContentT] | None:
        return self._node._right._node



# pylint: skip-file
__all__: Final[list[str]] = ["DatabaseConnector"]

from typing import Final

from nrw.database._query_result import QueryResult

class DatabaseConnector:

    __slots__: Final[tuple[str, str, str]] = (
        "_connection",
        "_current_query_result",
        "_message",
    )

    def __init__(
        self,
        ip: str,
        port: int,
        database: str,
        username: str,
        password: str,
    ) -> None: ...
    def execute_statement(self, sql_statement: str) -> None: ...
    @property
    def current_query_result(self) -> QueryResult | None: ...
    @property
    def error_message(self) -> str | None: ...
    def close(self) -> None: ...



"""Klasse `Client`."""

from __future__ import annotations

__all__: Final[list[str]] = ["Client"]

import socket
import threading
from abc import ABC, abstractmethod
from contextlib import suppress
from typing import TYPE_CHECKING, Final

if TYPE_CHECKING:
    from io import TextIOWrapper


class _SocketWrapper:
    __slots__: Final[tuple[str, str, str]] = ("_socket", "_to_server", "_from_server")

    def __init__(self, server_ip: str, server_port: int) -> None:
        try:
            self._socket: socket.socket | None = socket.socket(
                socket.AF_INET,
                socket.SOCK_STREAM,
            )
            self._socket.connect((server_ip, server_port))
            self._to_server: TextIOWrapper | None = self._socket.makefile(
                mode="w",
                encoding="utf-8",
            )
            self._from_server: TextIOWrapper | None = self._socket.makefile(
                mode="r",
                encoding="utf-8",
            )
        except OSError:
            self._socket = None
            self._to_server = None
            self._from_server = None

    def receive(self) -> str | None:
        if self._from_server is not None:
            with suppress(OSError, ValueError):
                received_line: str = self._from_server.readline().strip()
                return received_line if received_line else None
        return None

    def send(self, message: str) -> None:
        if self._to_server is not None:
            with suppress(OSError, ValueError):
                self._to_server.write(f"{message}\n")
                self._to_server.flush()

    def close(self) -> None:
        if self._socket is not None:
            with suppress(OSError):
                self._to_server.close()
                self._from_server.close()
                self._socket.shutdown(socket.SHUT_RDWR)
                self._socket.close()


class Client(ABC):
    """Objekte von Unterklassen der abstrakten Klasse `Client` ermöglichen
    Netzwerkverbindungen zu einem Server mittels TCP/IP-Protokoll.
    Nach Verbindungsaufbau können Zeichenketten (`Strings`) zum Server gesendet und von
    diesem empfangen werden, wobei der Nachrichtenempfang nebenläufig geschieht.
    Zur Vereinfachung finden Nachrichtenversand und -empfang zeilenweise statt, d. h.,
    beim Senden einer Zeichenkette wird ein Zeilentrenner ergänzt und
    beim Empfang wird dieser entfernt. Jede empfangene Nachricht wird einer
    Ereignisbehandlungsmethode übergeben, die in Unterklassen implementiert werden muss.
    Es findet nur eine rudimentäre Fehlerbehandlung statt, so dass z. B.
    Verbindungsabbrüche nicht zu einem Programmabbruch führen.
    Eine einmal unterbrochene oder getrennte Verbindung kann nicht reaktiviert werden.
    """

    __slots__: Final[tuple[str, str]] = ("_socket_wrapper", "_active")

    def __init__(self, server_ip: str, server_port: int) -> None:
        """Es wird eine Verbindung zum durch `server_ip` und `server_port`
        spezifizierten Server aufgebaut, so dass Daten (Zeichenketten) gesendet und
        empfangen werden können. Kann die Verbindung nicht hergestellt werden, kann der
        Client nicht zum Datenaustausch verwendet werden.
        """
        self._socket_wrapper: _SocketWrapper = _SocketWrapper(server_ip, server_port)
        if self._socket_wrapper._socket is not None:
            self._active: bool = True
        else:
            self._active = False
        message_handler_thread: threading.Thread = threading.Thread(target=self._run)
        message_handler_thread.start()

    def _run(self) -> None:
        message: str | None = None
        while self._active:
            message = self._socket_wrapper.receive()
            if message is not None:
                self.process_message(message)
            else:
                self.close()

    def send(self, message: str) -> None:
        """Die Nachricht `message` wird - um einen Zeilentrenner ergänzt - an den Server
        gesendet. Schlägt der Versand fehl, geschieht nichts.
        """
        if self._active:
            self._socket_wrapper.send(message)

    def close(self) -> None:
        """Die Verbindung zum Server wird getrennt und der Client kann nicht mehr
        verwendet werden. Ist `Client` bereits vor Aufruf der Methode in diesem Zustand,
        geschieht nichts.
        """
        if self._active:
            self._active = False
            self._socket_wrapper.close()

    @property
    def is_connected(self) -> bool:
        """Die Anfrage liefert den Wert `True`, wenn der Client mit dem Server
        aktuell verbunden ist. Ansonsten liefert sie den Wert `False`.
        """
        return self._active

    @abstractmethod
    def process_message(self, message: str) -> None:
        """Diese Methode wird aufgerufen, wenn der Client die Nachricht `message` vom
        Server empfangen hat. Der vom Server ergänzte Zeilentrenner wurde zuvor
        entfernt. Die Methode ist abstrakt und muss in einer Unterklasse der Klasse
        `Client` überschrieben werden, so dass auf den Empfang der Nachricht reagiert
        wird. Der Aufruf der Methode erfolgt nicht synchronisiert.
        """



"""Materialien zu den zentralen NRW-Abiturprüfungen im Fach Informatik ab 2018
implementiert in Python.
"""

from __future__ import annotations

__version__: Final[str] = "1.4.0"

from typing import Final



"""Implementation des generischen Protocols `ComparableContent[_T_contra]`."""

from __future__ import annotations

__all__: Final[list[str]] = ["ComparableContent", "ComparableContentT"]

from typing import Any, Final, Protocol, TypeVar, runtime_checkable

_T_contra = TypeVar("_T_contra", contravariant=True)


@runtime_checkable
class ComparableContent(Protocol[_T_contra]):
    """Das generische Protocol `ComparableContent[_T_contra]` legt die Methoden
    fest, über die Objekte verfügen müssen, die in einen binaeren Suchbaum
    (`BinarySearchTree`) eingefügt werden sollen.
    """

    def __eq__(self, other: object, /) -> bool: ...
    def __gt__(self, other: _T_contra, /) -> bool: ...
    def __lt__(self, other: _T_contra, /) -> bool: ...


ComparableContentT = TypeVar("ComparableContentT", bound=ComparableContent[Any])



"""Implementation der Klasse `Graph`."""

from __future__ import annotations

__all__: Final[list[str]] = ["Graph"]

from io import StringIO
from typing import TYPE_CHECKING, Final

from nrw.datastructures._list import List

if TYPE_CHECKING:
    from nrw.datastructures._edge import Edge
    from nrw.datastructures._vertex import Vertex


class Graph:
    """Die Klasse `Graph` stellt einen ungerichteten, kantengewichteten Graphen dar.
    Es können Knoten- und Kantenobjekte hinzugefügt und entfernt,
    flache Kopien der Knoten- und Kantenlisten des Graphen angefragt
    und Markierungen von Knoten und Kanten gesetzt und überprueft werden.
    Des Weiteren kann eine Liste der Nachbarn eines bestimmten Knoten,
    eine Liste der inzidenten Kanten eines bestimmten Knoten
    und die Kante von einem bestimmten Knoten zu einem
    anderen bestimmten Knoten angefragt werden.
    Abgesehen davon kann abgefragt werden, welches
    Knotenobjekt zu einer bestimmten ID gehört und ob der Graph leer ist.
    """

    __slots__: Final[tuple[str, str]] = ("_vertices", "_edges")
    __hash__ = None  # type: ignore[assignment]

    def __init__(self) -> None:
        """Ein Objekt vom Typ `Graph` wird erstellt.
        Der von diesem Objekt repraesentierte Graph ist leer.
        """
        self._vertices: List[Vertex] = List()
        self._edges: List[Edge] = List()

    def __repr__(self) -> str:
        return (
            f"{self.__class__.__name__}(vertices={self._vertices!r}, "
            f"edges={self._edges!r})"
        )

    def __str__(self) -> str:
        if self.is_empty:
            return f"{self.__class__.__name__}()"

        with StringIO() as buffer:
            buffer.write(f"{self.__class__.__name__}(")
            self._vertices.to_first()
            while self._vertices.has_access:
                current: Vertex | None = self._vertices.content
                assert current is not None
                buffer.write(f"{current} -> {self.get_neighbours(current)}, ")
                self._vertices.next()
            return f"{buffer.getvalue().rstrip(', ')})"

    @property
    def vertices(self) -> List[Vertex]:
        """Die Anfrage liefert eine neue Liste aller Knotenobjekte
        vom Typ `List[Vertex]`.
        """
        result: List[Vertex] = List()
        self._vertices.to_first()
        while self._vertices.has_access:
            result.append(self._vertices.content)
            self._vertices.next()
        return result

    @property
    def edges(self) -> List[Edge]:
        """Die Anfrage liefert eine neue Liste aller Kantenobjekte
        vom Typ `List[Edge]`.
        """
        result: List[Edge] = List()
        self._edges.to_first()
        while self._edges.has_access:
            result.append(self._edges.content)
            self._edges.next()
        return result

    def get_vertex(self, id_: str) -> Vertex | None:
        """Die Anfrage liefert das Knotenobjekt mit `id_` als ID.
        Ist ein solchen Knotenobjekt nicht im Graphen enthalten,
        wird `None` zurückgeliefert.
        """
        self._vertices.to_first()
        while self._vertices.has_access:
            if self._vertices.content.id == id_:
                return self._vertices.content
            self._vertices.next()
        return None

    def add_vertex(self, vertex: Vertex | None) -> None:
        """Der Auftrag fügt den Knoten `vertex` in den Graphen ein,
        sofern es noch keinen Knoten mit demselben ID-Eintrag
        wie `vertex` im Graphen gibt und `vertex` eine ID hat, welche nicht `None` ist.
        Ansonsten passiert nichts.
        """
        if vertex is None or vertex.id is None:
            return

        self._vertices.to_first()
        while self._vertices.has_access:
            if self._vertices.content.id == vertex.id:
                return
            self._vertices.next()

        self._vertices.append(vertex)

    def remove_vertex(self, vertex: Vertex) -> None:
        """Der Auftrag entfernt den Knoten `vertex` aus dem Graphen
        und löscht alle Kanten, die mit ihm inzident sind.
        Ist der Knoten `vertex` nicht im Graphen enthalten, passiert nichts.
        """
        self._edges.to_first()
        while self._edges.has_access:
            if vertex in self._edges.content.vertices:
                self._edges.remove()
            else:
                self._edges.next()

        self._vertices.to_first()
        while self._vertices.has_access and self._vertices.content is not vertex:
            self._vertices.next()

        if self._vertices.has_access:
            self._vertices.remove()

    def get_edge(self, vertex: Vertex, another_vertex: Vertex) -> Edge | None:
        """Die Anfrage liefert die Kante, welche die Knoten `vertex`
        und `another_vertex` verbindet, als Objekt vom Typ `Edge`.
        Ist der Knoten `vertex` oder der Knoten `another_vertex` nicht
        im Graphen enthalten oder gibt es keine Kante, die beide Knoten verbindet,
        so wird `None` zurückgeliefert.
        """
        self._edges.to_first()
        while self._edges.has_access:
            vertex1, vertex2 = self._edges.content.vertices
            if (vertex1 is vertex and vertex2 is another_vertex) or (
                vertex1 is another_vertex and vertex2 is vertex
            ):
                return self._edges.content
            self._edges.next()
        return None

    def add_edge(self, edge: Edge | None) -> None:
        """Der Auftrag fügt die Kante `edge` in den Graphen ein,
        sofern beide durch die Kante verbundenen Knoten im Graphen enthalten sind,
        nicht identisch sind und noch keine Kante zwischen den Knoten existiert.
        Ansonsten passiert nichts.
        """
        if edge is None:
            return

        vertex1, vertex2 = edge.vertices
        # pylint: disable=R0916
        if (
            vertex1 is not None
            and vertex2 is not None
            and self.get_vertex(vertex1.id) is vertex1
            and self.get_vertex(vertex2.id) is vertex2
            and self.get_edge(vertex1, vertex2) is None
            and vertex1 is not vertex2
        ):
            self._edges.append(edge)

    def remove_edge(self, edge: Edge) -> None:
        """Der Auftrag entfernt die Kante `edge` aus dem Graphen.
        Ist die Kante `edge` nicht im Graphen enthalten, passiert nichts.
        """
        self._edges.to_first()
        while self._edges.has_access:
            if self._edges.content is edge:
                self._edges.remove()
                return
            self._edges.next()

    def set_all_vertex_marks(self, mark: bool) -> None:
        """Der Auftrag setzt die Markierungen aller Knoten des Graphen auf `mark`."""
        self._vertices.to_first()
        while self._vertices.has_access:
            self._vertices.content.mark = mark
            self._vertices.next()

    def all_vertices_marked(self) -> bool:
        """Die Anfrage liefert `True`,
        wenn alle Knoten des Graphen mit `True` markiert sind, ansonsten `False`.
        """
        self._vertices.to_first()
        while self._vertices.has_access:
            if not self._vertices.content.is_marked:
                return False
            self._vertices.next()
        return True

    def set_all_edge_marks(self, mark: bool) -> None:
        """Der Auftrag setzt die Markierungen aller Kanten des Graphen auf `mark`."""
        self._edges.to_first()
        while self._edges.has_access:
            self._edges.content.mark = mark
            self._edges.next()

    def all_edges_marked(self) -> bool:
        """Die Anfrage liefert `True`,
        wenn alle Kanten des Graphen mit `True` markiert sind, ansonsten `False`.
        """
        self._edges.to_first()
        while self._edges.has_access:
            if not self._edges.content.mark:
                return False
            self._edges.next()
        return True

    def get_neighbours(self, vertex: Vertex) -> List[Vertex]:
        """Die Anfrage liefert alle Nachbarn des Knotens `vertex`
        als neue Liste vom Typ `List[Vertex]`.
        Hat der Knoten `vertex` keine Nachbarn in diesem Graphen
        oder ist gar nicht in diesem Graphen enthalten,
        so wird eine leere Liste zurückgeliefert.
        """
        result: List[Vertex] = List()

        self._edges.to_first()
        while self._edges.has_access:
            vertex1, vertex2 = self._edges.content.vertices
            if vertex is vertex1:
                result.append(vertex2)
            elif vertex is vertex2:
                result.append(vertex1)
            self._edges.next()
        return result

    def get_edges(self, vertex: Vertex) -> List[Edge]:
        """Die Anfrage liefert eine neue Liste alle inzidenten Kanten
        zum Knoten `vertex`.
        Hat der Knoten `vertex` keine inzidenten Kanten in diesem Graphen
        oder ist gar nicht in diesem Graphen enthalten,
        so wird eine leere Liste zurückgeliefert.
        """
        result: List[Edge] = List()

        self._edges.to_first()
        while self._edges.has_access:
            if vertex in self._edges.content.vertices:
                result.append(self._edges.content)
            self._edges.next()
        return result

    @property
    def is_empty(self) -> bool:
        """Die Anfrage liefert `True`, wenn der Graph keine Knoten enthält,
        ansonsten `False`.
        """
        return self._vertices.is_empty



"""Traversierung für Binär Bäume."""

from __future__ import annotations

__all__: Final[list[str]] = [
    "preorder",
    "inorder",
    "postorder",
    "levelorder",
    "reverse_inorder",
]

from typing import Final, TypeVar
from warnings import warn

from nrw.datastructures import BinarySearchTree, BinaryTree, ComparableContentT, List

_T = TypeVar("_T")


def preorder(
    tree: BinaryTree[_T] | BinarySearchTree[ComparableContentT],
    *,
    reverse: bool = False,
) -> List[_T | ComparableContentT]:
    result: List[_T | ComparableContentT] = List()

    if tree.is_empty:
        return result

    if not reverse:
        result.append(tree.content)
        result.concat(preorder(tree.left_tree, reverse=reverse))
        result.concat(preorder(tree.right_tree, reverse=reverse))
    else:
        result.append(tree.content)
        result.concat(preorder(tree.right_tree, reverse=reverse))
        result.concat(preorder(tree.left_tree, reverse=reverse))

    return result


def inorder(
    tree: BinaryTree[_T] | BinarySearchTree[ComparableContentT],
    *,
    reverse: bool = False,
) -> List[_T | ComparableContentT]:
    result: List[_T | ComparableContentT] = List()

    if tree.is_empty:
        return result

    if not reverse:
        result.concat(inorder(tree.left_tree, reverse=reverse))
        result.append(tree.content)
        result.concat(inorder(tree.right_tree, reverse=reverse))
    else:
        result.concat(inorder(tree.right_tree, reverse=reverse))
        result.append(tree.content)
        result.concat(inorder(tree.left_tree, reverse=reverse))

    return result


def reverse_inorder(
    tree: BinaryTree[_T] | BinarySearchTree[ComparableContentT],
) -> List[_T | ComparableContentT]:  # pragma: no cover
    warn("Use 'inorder(..., reverse=True)'", DeprecationWarning, stacklevel=2)
    return inorder(tree, reverse=True)


def postorder(
    tree: BinaryTree[_T] | BinarySearchTree[ComparableContentT],
    *,
    reverse: bool = False,
) -> List[_T | ComparableContentT]:
    result: List[_T | ComparableContentT] = List()

    if tree.is_empty:
        return result

    if not reverse:
        result.concat(postorder(tree.left_tree, reverse=reverse))
        result.concat(postorder(tree.right_tree, reverse=reverse))
        result.append(tree.content)
    else:
        result.concat(postorder(tree.right_tree, reverse=reverse))
        result.concat(postorder(tree.left_tree, reverse=reverse))
        result.append(tree.content)

    return result


def levelorder(
    tree: BinaryTree[_T] | BinarySearchTree[ComparableContentT],
    *,
    reverse: bool = False,
) -> List[_T | ComparableContentT]:
    if tree.is_empty:
        return List()

    trees: List[BinaryTree[_T] | BinarySearchTree[ComparableContentT]] = List()
    trees.append(tree)
    trees.to_first()
    while trees.has_access:
        current_tree: BinaryTree[_T] | BinarySearchTree[ComparableContentT] | None = (
            trees.content
        )
        assert current_tree is not None

        if not reverse:
            if not current_tree.left_tree.is_empty:
                trees.append(current_tree.left_tree)
            if not current_tree.right_tree.is_empty:
                trees.append(current_tree.right_tree)
        else:
            if not current_tree.right_tree.is_empty:
                trees.append(current_tree.right_tree)
            if not current_tree.left_tree.is_empty:
                trees.append(current_tree.left_tree)
        trees.next()

    result: List[_T | ComparableContentT] = List()
    trees.to_first()
    while trees.has_access:
        result.append(trees.content.content)
        trees.next()
    return result



"""Implementation der Klasse `Edge`."""

from __future__ import annotations

__all__: Final[list[str]] = ["Edge"]

from typing import TYPE_CHECKING, Final

if TYPE_CHECKING:
    from nrw.datastructures._vertex import Vertex


class Edge:
    """Die Klasse `Edge` stellt eine einzelne, ungerichtete Kante eines Graphen dar.
    Beim Erstellen werden die beiden durch sie zu verbindenden Knotenobjekte und eine
    Gewichtung als `int` übergeben. Beide Knotenobjekte können abgefragt werden.
    Des Weiteren koennen die Gewichtung
    und eine Markierung gesetzt und abgefragt werden.
    """

    __slots__: Final[tuple[str, str, str]] = ("_vertices", "_weight", "_mark")
    __hash__ = None  # type: ignore[assignment]

    def __init__(self, vertex: Vertex, another_vertex: Vertex, weight: int) -> None:
        """Ein neues Objekt vom Typ `Edge` wird erstellt. Die von diesem Objekt
        repraesentierte Kante verbindet die Knoten `vertex` und `another_vertex` mit der
        Gewichtung `weight`. Ihre Markierung hat den Wert `False`.
        """
        self._vertices: tuple[Vertex, Vertex] = (vertex, another_vertex)
        self._weight: int = weight
        self._mark: bool = False

    def __repr__(self) -> str:
        return (
            f"{self.__class__.__name__}(vertices={self._vertices!r}, "
            f"weight={self._weight!r}, mark={self._mark!r})"
        )

    def __str__(self) -> str:
        vertex1, vertex2 = self._vertices
        return f"{vertex1} --{self._weight}-- {vertex2}"

    @property
    def vertices(self) -> tuple[Vertex, Vertex]:
        """Die Anfrage gibt die beiden Knoten, die durch die Kante verbunden werden,
        als `tuple` vom Typ `Vertex` zurück.
        Das `tuple` hat genau zwei Einträge mit den Indexwerten 0 und 1.
        """
        return self._vertices

    @property
    def weight(self) -> int:
        """Der Auftrag setzt das Gewicht der Kante auf `weight`."""
        return self._weight

    @weight.setter
    def weight(self, new_weight: int) -> None:
        self._weight = new_weight

    @property
    def mark(self) -> bool:
        """Die Anfrage liefert `True`,
        wenn die Markierung der Kante den Wert `True` hat, ansonsten `False`.
        """
        return self._mark

    @mark.setter
    def mark(self, new_mark: bool) -> None:
        self._mark = new_mark

    @property
    def is_marked(self) -> bool:
        """Die Anfrage liefert `True`,
        wenn die Markierung der Kante den Wert `True` hat, ansonsten `False`.
        """
        return self._mark
